#include "main.h"
#include "wav.h"
#include "GymReaderCpp.h"
#include "ym3438.h"

const unsigned SAMPLE_RATE = 7670454 / 144;

ym3438_t ym3438;
short ym3438_accm[24][2];
int ym3438_sample[2];
int ym3438_cycles = 0;
unsigned ym3438_processed_samples = 0;
vector<sample> output(2 * SAMPLE_RATE * 10); // 2=stereo

static void doYmClock() {
    short buffer[2];
    OPN2_Clock(&ym3438, buffer);
    //OPN2_Clock(&ym3438, ym3438_accm[ym3438_cycles]);
    //ym3438_cycles = (ym3438_cycles + 1) % 24;
    //if (ym3438_cycles == 0)
    //{
    //    ym3438_sample[0] = 0;
    //    ym3438_sample[1] = 0;
    //    for (unsigned j = 0; j < 24; j++)
    //    {
    //        ym3438_sample[0] += ym3438_accm[j][0];
    //        ym3438_sample[1] += ym3438_accm[j][1];
    //    }
    //}
    if (2 * ym3438_processed_samples >= output.size()) {
        printf("Too many samples\n");
        return;
    }
    //output[2 * ym3438_processed_samples] = ym3438_sample[0] * 11;
    //output[2 * ym3438_processed_samples + 1] = ym3438_sample[1] * 11;
    output[2 * ym3438_processed_samples] = buffer[0] * 11;
    output[2 * ym3438_processed_samples + 1] = buffer[1] * 11;
    ym3438_processed_samples += 1;
}

struct GymReader {
    ym3438_t& ym3438;

    GymReader(const char *fileName, ym3438_t& ym3438) : ym3438(ym3438) {
        std::ifstream f(fileName, std::ios::binary);
        file.assign(std::istreambuf_iterator<char>(f), std::istreambuf_iterator<char>());
        fileOffset = 428;
    }

    bool processInstruction() {
        const unsigned waitStates = 24;
        if (fileOffset < file.size()) {
            switch (file[fileOffset]) {
            case 0:
                fileOffset += 1;
                return true;
            case 1:
                printf("YM port0 write %02x [%02x]\n", file[fileOffset + 1], file[fileOffset + 2]);
                OPN2_Write(&ym3438, 0, file[fileOffset + 1]); // YM2612 A0 (0x4000)
                for (unsigned i = 0; i < waitStates; i++) doYmClock();
                OPN2_Write(&ym3438, 1, file[fileOffset + 2]); // YM2612 D0 (0x4001)
                for (unsigned i = 0; i < waitStates; i++) doYmClock();
                fileOffset += 3;
                return false;
            case 2:
                printf("YM port1 write %02x [%02x]\n", file[fileOffset + 1], file[fileOffset + 2]);
                OPN2_Write(&ym3438, 2, file[fileOffset + 1]); // YM2612 A1 (0x4002)
                for (unsigned i = 0; i < waitStates; i++) doYmClock();
                OPN2_Write(&ym3438, 3, file[fileOffset + 2]); // YM2612 D1 (0x4003)
                for (unsigned i = 0; i < waitStates; i++) doYmClock();
                fileOffset += 3;
                return false;
            case 3:
                printf("PSG data=%02x\n", file[fileOffset + 1]);
                fileOffset += 2;
                return false;
            default:
                printf("Unknown command %02x\n", file[fileOffset]);
                return false;
            }
        }
        printf("End of file\n");
        return true;
    }

private:
    vector<byte> file;
    unsigned fileOffset;
};


void run() {
    const unsigned SAMPLES_PER_FRAME = SAMPLE_RATE / 60;

    OPN2_Reset(&ym3438);
    OPN2_SetChipType(ym3438_mode_ym2612);
    memset(ym3438_accm, 0, sizeof(ym3438_accm));
    memset(ym3438_sample, 0, sizeof(ym3438_sample));

    GymReader reader("file.gym", ym3438);
    unsigned processedFrames = 0;

    while (true) {
        unsigned target = ym3438_processed_samples + SAMPLES_PER_FRAME;
        bool frameComplete = false;

        while (ym3438_processed_samples < target) {
            if (!frameComplete) {
                frameComplete = reader.processInstruction();
            }
            doYmClock();
            if (2 * ym3438_processed_samples >= output.size()) goto after;
        }
        processedFrames += 1;
    }

after:
    writeWavFile("example.wav", output, false);
    printf("Processed %d frames\n", processedFrames);
}



int main() {
    vector<short> wave(44100);
    auto frequency = 5000;
    auto SAMPLE_RATE = 44100;
    double angle = 0;

    run();

    //for (int i = 0; i < 44100; i++) {
    //    wave[i] = (short)(sin(angle) * 2000);
    //    angle += frequency * 2 * M_PI / SAMPLE_RATE;
    //}
    //
    //writeWavFile("example.wav", wave, true);
}
